require 'rails_helper'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  require_relative '../../swagger_helper'

  ACCEPTABLE_PARAMS = %w[
    nud
    sne_profile_creation_date
    ss_id
    first_name
    last_name
    patronymic_name
    nationality
    email
    email_advances_assistant
    event_notification
    civility
    birth_date
    work_contract_type
    profession_details
    has_other_employers
    main_employer_name
    main_employer_siret
    workplace_postal_code
    workplace_locality
    foreign_workplace_postal_code
    foreign_workplace_locality
    foreign_workplace_country
    year_minus_one
    amount_year_minus_one
    year_minus_two
    amount_year_minus_two
    ressources_details
    tax_revenues
    apl_amount
    postal_details
    actual_housing_details
    family_situation
    family_situation_validated
    asked_children_nb
    valid_children_nb
    situation_tax_treshold
    status
    updated_status_at
    asked_locataire_hlm
    valid_locataire_hlm
    valid_priority_public
    asked_priority_public
    single_parent
    priority_public
    mobility
    mutation
    completed_requests_nb
    on_going_request_nb
    disability
    renewal_date
    eligibility_query
    active_housing_request_id
    housing_request_ids
    entity_id
    application_ids
    monthly_incomes
    monthly_charges
    cellphone
    sms_notification
    homephone
    workphone
    main_user_id
    unemployment_period
    command_to_leave_date
    pro_or_home_tel_number
    unemployment_period_date_end
  ]

  MUTATE_PARAMS = {
    nud: 'String',
    sne_profile_creation_date: 'DateTime',
    ss_id: 'String',
    first_name: 'String',
    last_name: 'String',
    patronymic_name: 'String',
    nationality: 'String',
    email: 'String',
    email_advances_assistant: 'String',
    event_notification: 'Boolean',
    civility: 'String',
    birth_date: 'Date',
    work_contract_type: 'String',
    profession_details: 'Hash',
    has_other_employers: 'Boolean',
    main_employer_name: 'String',
   # main_employer_siret: 'String',
    workplace_postal_code: 'String',
    workplace_locality: 'String',
    foreign_workplace_postal_code: 'String',
    foreign_workplace_locality: 'String',
    foreign_workplace_country: 'String',
    year_minus_one: 'String',
    amount_year_minus_one: 'Float',
    year_minus_two: 'String',
    amount_year_minus_two: 'Float',
    ressources_details: 'Hash',
    tax_revenues: 'Float',
    apl_amount: 'Float',
    postal_details: 'Hash',
    actual_housing_details: 'Hash',
    family_situation: 'String',
    family_situation_validated: 'Boolean',
    situation_tax_treshold: 'Integer',
    status: 'String',
    asked_locataire_hlm: 'Boolean',
    valid_locataire_hlm: 'Boolean',
    valid_priority_public: 'Boolean',
    asked_priority_public: 'Boolean',
    priority_public: 'Boolean',
    mobility: 'Boolean',
    mutation: 'Boolean',
    disability: 'String',
    renewal_date: 'DateTime',
    monthly_incomes: 'Hash',
    monthly_charges: 'Hash',
    cellphone: 'String',
    is_owner: 'Boolean',
    'sms_notification': 'Boolean',
    'homephone': "String",
    'workphone': 'String',
    main_user_id: 'String',
    workplace_insee_code: "String",
    unemployment_period: "Boolean",
    command_to_leave_date: "Date",
    pro_or_home_tel_number: "String",
    unemployment_period_date_end: "Date"
  }

  OFFERS_ACCEPTABLE_PARAMS = %w[
    created_at
    updated_at
    eligibility_query
    typology
    lessor_status
    type
    age
    district
    department
    publication_end_date
    rent_amount
    surface
    rooms
    exclusivity_end_date
    nb_of_eligible_employees
    nb_of_requests_from_employees
    availability_date
    description
    details
    old_status
    status
    status_updated_at
    effort_rate
    min_effort_treshold
    single_parent
    reserved
    priority_public
    mobility
    bedrooms
    bathrooms
    heating_type
    floors
    has_elevator
    has_gardian
    dpe
    outside
    has_cellar
    parking_type
    pmr
    kitchen_type
    address
    postal_code
    country
    latitude
    longitude
    mutation
    eligibility_string
    applicant_category_1
    threshold_type
    threshold_subtype
    threshold_category_2
    threshold_category_3
    threshold_category_4
    threshold_category_5
    threshold_category_6
    threshold_category_7
    threshold_category_8
    threshold_category_9
    threshold_category_10
    threshold_category_11
    threshold_category_12
    individual_heating
    date_publication_start
    date_publication_end
    lessor_id
    kind
    building_type
    allready_rent
    guarantee_deposit
    rental_charges
    car_park_rent
    offer_status
    offer_status_updated_at
    lessor_status_updated_at
    ceil_category_1
    ceil_category_2
    ceil_category_3
    ceil_category_4
    ceil_category_5
    ceil_category_6
    ceil_category_7
    ceil_category_8
    ceil_category_9
    ceil_category_10
    ceil_category_11
    ceil_category_12
    floor_category_1
    floor_category_2
    floor_category_3
    floor_category_4
    floor_category_5
    floor_category_6
    floor_category_7
    floor_category_8
    floor_category_9
    floor_category_10
    floor_category_11
    floor_category_12
    min_revenues
    total_charges
    max_effort_rate
    charges
    tense_area
    subsidiary_company
    rpls_id
    operation_name
    administrator
    administration_area
    potential_request_name_town
    potential_request_name_county
    number_of_companies_30_minutes_around
    construction_year
    address_can_be_shared
    qpv
    total_floors
    address_extended
    dpe_gas
    test_created_at
    new_offer
    lessor_name
    lessor_siren
    als_siren
    postal_code
    residence_code
    pictures
    agency_id
    agency_name
    lessor_contact_name
    lessor_email
    lessor_phone_number
    garden
    green_area
    condo
    furnished_lodging
    virtual_visit
    duplex
    first_quartile
    other_quartiles
    lessor_convention_reference
    reserving_agency_name
    reserving_agency_siren
    is_owner
  ]

  VALID_OPERATORS = %w[
    $in
    $nin
    $ne
    $gt
    $gte
    $lt
    $lte
    $eq
    $regex
    $all
  ]

  describe Api::Dmo::ApplicantsController, :type => :api do
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper

    describe 'with project oriented roles' do
      let!(:demandeur) { create(:user, auth_roles: ['demandeur']) }

      let!(:bailleur) { create(:user, auth_roles: ['bailleur']) }
      let!(:entreprise) { create(:user, auth_roles: ['entreprise']) }
      let!(:dc_channel) {create(:housing_channel, key: 'dc')}

      describe 'POST as entreprise with main_user_id' do
        before :each do
          authenticated_header(request, demandeur)
          post 'create', {
            data: {
              attributes: {
                main_user_id: demandeur.id.to_s
              }
            }
          }
        end

        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end

        it 'returns valid json' do
          attributes = json['data']['attributes']
          expect(attributes['main_user_id']).to eq demandeur.id.to_s
        end
      end

    end

    describe 'normal' do
      let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
      let!(:superadministrateur) {create(:user, auth_roles: ['superadministrateur'])}
      let!(:user) {create(:user, first_name: 'Jackie', last_name: 'Mike', email: 'jackie.mike@gmail.com')}

      describe 'POST /applicants' do
        describe 'applicant created with identity' do
          before :each do
            authenticated_header(request, user)
            post 'create', {
              data: {
                attributes: {
                  first_name: 'Joseph',
                  last_name: 'Joestar',
                  email: 'josephjoestar@test.fr'
                }
              }
            }
          end

          it 'returns status code 201' do
            expect(response).to have_http_status(201)
          end

          it 'returns valid JSON' do
            attributes = json['data']['attributes']
            expect(attributes['full_name']).to eq 'Joseph Joestar'
            expect(attributes['email']).to eq 'josephjoestar@test.fr'
          end
        end

        describe 'Create with nud' do
          let!(:lessor_user) { create(:user, :lessor) }

          before :each do
            authenticated_header(request, lessor_user)
          end

          it 'returns valid JSON' do
            post 'create', {
              data: {
                attributes: {
                  nud: '0130719003263G3U05'
                }
              }
            }
            attributes = json['data']['attributes']
          end
        end

        describe 'Create applicant with minimal valid datas' do
          before :each do
            authenticated_header(request, responsible)
            post 'create', {
              data: {
                attributes: {
                  main_user_id: responsible.id,
                }
              }
            }
          end
          it 'returns status code 201' do
            expect(response).to have_http_status(201)
          end

          describe 'then same user creates another applicant' do
            before :each do
              authenticated_header(request, responsible)
              post 'create', {
                data: {
                  attributes: {
                    first_name: 'Janine',
                    last_name: 'Joseph',
                    email: 'janinejoseph@test.fr'
                  }
                }
              }
            end

            it 'returns status code 201' do
              expect(response).to have_http_status(201)
            end

            it 'returns valid JSON' do
              attributes = json['data']['attributes']
              expect(attributes['full_name']).to eq 'Janine Joseph'
              expect(attributes['email']).to eq 'janinejoseph@test.fr'
            end
          end
        end
        MUTATE_PARAMS.each do |param, type|
          describe 'post request with ' + param.to_s + ' ' + type.to_s + ' is valid' do
            let(:params) do
              attr = {}
              attr[param] = get_typed_value(type)
              attr
            end
            before :each do
              authenticated_header(request, responsible)
              post 'create', {
                  data: {
                      attributes: params
                  }
              }
            end
            it 'returns status code 201 and the value appears in json' do
              expect(response).to have_http_status(201)
              if (type == 'Date' || type == 'DateTime')
                expect(json['data']['attributes'][param.to_s]).not_to be_empty
                expect(json['data']['attributes'][param.to_s]).not_to eq('')
              else
                expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type))
              end
            end
          end
        end

        # post with housing_request_ids:
        describe 'post request with housing_request_ids' do
          let!(:applicant2) {create(:applicant)}
          let!(:housing_request) {create(:housing_request, main_applicant: applicant2, is_active: true)}
          let!(:housing_request2) {create(:housing_request, main_applicant: applicant2)}
          before :each do
            authenticated_header(request, responsible)
            post 'create', {
                data: {
                    attributes: {
                        housing_request_ids: [housing_request.id.to_s, housing_request2.id.to_s],
                        first_name: 'Michel',
                        last_name: "Dupont",
                        email: "michel.dupont@gmail.com"
                    }
                }
            }
          end
          it 'returns status code 201 and the value appears in json' do
            expect(response).to have_http_status(201)
            #TODO on pourrait s'attendre a retrouver les ids des housings requests mais c'est coherent car elles sont deja en belongs to avec un autre applicant
            expect(json['data']['attributes']['housing_request_ids']).to eq([])
          end

          describe 'post request with procurating user' do
            let!(:applicant2) {create(:applicant)}
            let!(:housing_request) {create(:housing_request, main_applicant: applicant2, is_active: true)}
            let!(:housing_request2) {create(:housing_request, main_applicant: applicant2)}
            before :each do
              authenticated_header(request, responsible)
              post 'create', {
                  data: {
                      attributes: {
                          first_name: 'Michel',
                          last_name: "Dupont",
                          email: "michel.dupont@gmail.com"
                      }
                  }
              }
            end
            it 'returns status code 201 and the value appears in json' do
              expect(response).to have_http_status(201)
              #TODO on pourrait s'attendre a retrouver les ids des housings requests mais c'est coherent car elles sont deja en belongs to avec un autre applicant
              expect(json['data']['attributes']['first_name']).to eq('Michel')
              expect(json['data']['attributes']['writers']).to eq([responsible.id.to_s])
            end
          end
        end
      end

      describe 'GET /applicants' do
        let!(:applicant1) {create(:applicant, nationality: 'belge', apl_amount: 50)}
        let!(:applicant3) {create(:applicant, nationality: 'quebecquoise', apl_amount: 150)}
        let!(:applicant2) {create(:applicant, nationality: 'francaise', apl_amount: 150)}

        [
          {user: 'responsible', total_objects: 3},
          {user: 'superadministrateur', total_objects: 3},
          {user: 'user', total_objects: 0},
        ].each do |a|
          describe "as #{a[:user]}" do

            before :each do
              authenticated_header(request, eval(a[:user]))
              get 'index'
            end
            it 'returns status code 200' do
              expect(response).to have_http_status(200)
            end
            it 'returns a valid json' do
              expect(json).not_to be_empty
              expect(json['data'].length).to eq(a[:total_objects])
              expect(json['meta']['pagination']).to be_present
              expect(json['meta']['pagination']['total_objects']).to eq(a[:total_objects])
            end
          end
        end

        ACCEPTABLE_PARAMS.each do |acceptable_param|
          type = MUTATE_PARAMS[acceptable_param.to_sym] || MUTATE_PARAMS[acceptable_param.to_s]
          if type
            describe 'GET /applicants with sort' do
              let!(:applicant3) {create(:applicant, auto_set_user: false, acceptable_param => get_typed_value(type, 2))}
              let!(:applicant1) {create(:applicant, auto_set_user: false, acceptable_param => get_typed_value(type, 0))}
              let!(:applicant2) {create(:applicant, auto_set_user: false, acceptable_param => get_typed_value(type, 1))}
              before :each do
                authenticated_header(request, responsible)
                get 'index', {sort: {acceptable_param => 1}}
              end
              it 'returns sorted results' do
                expect(response.status).to eq(200)
                if type == 'DateTime'
                  applicant_attribute = json['data'].map {|unit| DateTime.parse(unit['attributes'][acceptable_param])}
                  expect(applicant_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
                elsif type == 'Date'
                  applicant_attribute = json['data'].map {|unit| Date.parse(unit['attributes'][acceptable_param])}
                  expect(applicant_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
                else
                  applicant_attribute = json['data'].map {|unit| unit['attributes'][acceptable_param]}
                  expect(applicant_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
                end
              end
            end
          end
        end

        describe 'GET /applicants with operators' do
          let!(:applicant1) {create(:applicant, nationality: 'belge', apl_amount: 50)}
          let!(:applicant3) {create(:applicant, nationality: 'quebecquoise', apl_amount: 150)}
          let!(:applicant2) {create(:applicant, nationality: 'Samoan', apl_amount: 150)}
          describe 'with $lt' do
            before :each do
              authenticated_header(request, responsible)
              get 'index', {apl_amount: {'$lt': 100}}
            end
            it 'returns results' do
              expect(response.status).to eq(200)
              expect(json['meta']['pagination']['total_objects']).to eq(1)
            end
          end
          describe 'with $in' do
            before :each do
              authenticated_header(request, responsible)
              get 'index', {nationality: {'$in': ['belge', 'flamand']}}
            end
            it 'returns results' do
              expect(response.status).to eq(200)
              expect(json['meta']['pagination']['total_objects']).to eq(1)
            end
          end
          describe 'with $ne' do
            before :each do
              authenticated_header(request, responsible)
              get 'index', {apl_amount: {'$ne': 50}}
            end
            it 'returns sorted results' do
              expect(response.status).to eq(200)
              expect(json['meta']['pagination']['total_objects']).to eq(2)
            end
          end
          describe 'with $regex' do
            before :each do
              authenticated_header(request, responsible)
              get 'index', {nationality: {'$regex': /^S/}}
            end
            it 'returns sorted results' do
              expect(response.status).to eq(200)
              expect(json['meta']['pagination']['total_objects']).to eq(1)
            end
          end
        end
      end

      describe 'GET /applicants/:id' do
        describe 'verif it work' do
          let!(:applicant1) {create(:applicant, nationality: 'belge', apl_amount: 50)}

          # it_behaves_like 'true user only collaborator allowed' do
          #   let(:applicant) { applicant1 }
          #   let(:path) { ['show', { id: applicant1.id }] }
          # end

          before :each do
            authenticated_header(request, responsible)
            get 'show', {id: applicant1.id}
          end
          it 'returns status code 200' do
            expect(response).to have_http_status(200)
          end
          it 'returns a valid json' do
            expect(json).not_to be_empty
          end
        end
        describe 'verif it work not work with a non authorized user' do
          let!(:applicant1) {create(:applicant, nationality: 'belge', apl_amount: 50)}

          before :each do
            authenticated_header(request, user)
            get 'show', {id: applicant1.id}
          end
          it 'returns status code 404' do
            expect(response).to have_http_status(404)
          end
        end

        describe 'verif it work  for superadministrateur' do
          let!(:applicant1) {create(:applicant, nationality: 'belge', apl_amount: 50)}

          before :each do
            authenticated_header(request, superadministrateur)
            get 'show', {id: applicant1.id}
          end
          it 'returns status code 200' do
            expect(response).to have_http_status(200)
          end
        end
      end

      describe 'PATCH /applicant/:id for year_minus_two' do
        let!(:demandeur_test) { create(:user, auth_roles: ['demandeur']) }
        let!(:applicant_test) do
          create(
            :applicant,
            amount_year_minus_two: 1200,
            main_user_id: demandeur_test.id,
            user_id: demandeur_test.id,
            monthly_incomes: {
              'SAL': 1234
            }
          )
        end

        let!(:request_test) do
          create(
            :housing_request,
            is_active: true,
            main_applicant: applicant_test
          )
        end

        let(:params) { { amount_year_minus_two: 1201.0 } }

        before :each do
          authenticated_header(request, demandeur_test)
          patch 'update', {
            id: applicant_test.id.to_s,
            data: {
              attributes: params
            }
          }
        end

        it 'returns 200' do
          expect(response).to have_http_status(200)
        end

        it 'returns valid json' do
          data = json['data']
          attributes = data['attributes']
          expect(data['id']).to eq applicant_test.id.to_s
          expect(attributes['eligibility_query']).to_not eq "{is_public:true}"
          expect(attributes['amount_year_minus_two']).to be params[:amount_year_minus_two]
          expect(attributes['eligibility_query'].json_to_h[:ceil_category_1]['$gte']).to be 1200.0
        end

        describe 'GET same object' do
          before :each do
            authenticated_header(request, demandeur_test)
            get 'show', id: applicant_test.id.to_s
          end

          it 'returns 200' do
            expect(response).to have_http_status(200)
          end

          it 'returns valid json' do
            data = json['data']
            attributes = data['attributes']
            expect(data['id']).to eq applicant_test.id.to_s
            expect(attributes['eligibility_query']).to_not eq "{is_public:true}"
            expect(attributes['amount_year_minus_two']).to be params[:amount_year_minus_two]
            expect(attributes['eligibility_query'].json_to_h[:ceil_category_1]['$gte']).to be params[:amount_year_minus_two]
          end
        end

        describe 'update_forbidden if application but allowed for admins' do
          let!(:housing_offer1) {create(:housing_offer)}
          let!(:housing_application2) {create(:housing_application, housing_request: request_test, housing_offer: housing_offer1)}
          let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}

          before :each do
            authenticated_header(request, responsible)
            patch 'update', {
              id: applicant_test.id.to_s,
              data: {
                attributes: { amount_year_minus_two: 11.0 }
              }
            }
          end

          it 'returns 200' do
            expect(response).to have_http_status(200)
          end

        end
      end

      describe 'PATCH /applicants/:id' do
        applicant_params = {}
        MUTATE_PARAMS.each do |param, type|
          applicant_params[param] = get_typed_value(type)
        end
        let!(:applicant) {create(:applicant, **applicant_params)}

        MUTATE_PARAMS.each do |param, type|
          describe 'patch request with ' + param.to_s + ' ' + type.to_s + 'is valid' do
            let(:params) do
              attr = {}
              attr[param] = get_typed_value(type, 2)
              attr
            end
            before :each do
              authenticated_header(request, responsible)
              patch 'update', {
                id: applicant.id.to_s,
                data: {
                  attributes: params
                }
              }
            end
            it 'returns status code 200 and the value appears in json' do
              expect(response).to have_http_status(200)
              if (type == 'Date' || type == 'DateTime')
                expect(json['data']['attributes'][param.to_s]).not_to be_empty
                expect(json['data']['attributes'][param.to_s]).not_to eq('')
              else
                expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type, 2))
              end
            end
          end
        end
        describe 'put' do
          let!(:applicant3) {create(:applicant, auto_set_user: false, first_name: 'Jac', last_name: "Mike", email: "jackie.mike@gmail.com", writers:[responsible.id], status: "OK", monthly_incomes: {porn: 30}, monthly_charges: {cat_food: 10})}
          before :each do
            authenticated_header(request, responsible)
            patch 'update', {
              id: applicant3.id.to_s,
              data: {
                attributes: {
                  sms_notification: true
                }
              }
            }
          end
          it 'returns status code 200 and the value appears in json' do
            expect(response).to have_http_status(200)
            expect(json['data']['attributes']['first_name']).to eq("Jac")
          end
        end


        describe 'patch applicant entity_id as applicant' do
          let!(:entity) {create(:entity, siret: '00000000000001')}
          let!(:entity2) {create(:entity, siret: '00000000000002')}
          let!(:user1){create(:user, auth_roles: ["demandeur"])}
          let!(:reader_entity_1){create(:user, group_ids: [entity.manager_group_id])}
          let!(:reader_entity_2){create(:user, group_ids: [entity2.manager_group_id])}
          let!(:applicant){create(:applicant, user: user1 )}
          let!(:hr) {create(:housing_request, main_applicant: applicant, entity_validated: true, is_active: true, user: user1)}
          before :each do
            authenticated_header(request, user1)
            patch 'update', {
              id: applicant.id.to_s,
              data: {
                attributes: {
                  main_employer_siret: entity2.siret,
                  readers: [reader_entity_1.id], # Les readers et writers ne doivent pas être enregistré par le front
                  writers: [reader_entity_1.id], #
                }
              }
            }
          end
          it 'returns status code 200 and the value appears in json' do
            expect(response).to have_http_status(200)
            expect(applicant.reload.readable_by?(user1)).to eq true
            expect(hr.reload.entity).to eq(entity2)
            expect(hr.reload.entity_validated).to eq(false)
            expect(hr.reload.readable_by?(user1)).to eq true
            expect(hr.reload.readable_by?(reader_entity_1)).to eq false
            expect(hr.reload.readable_by?(reader_entity_2)).to eq true
          end
        end
      end

      describe 'DELETE /applicants/:id' do
        let!(:applicant) {create(:applicant, user_id: responsible.id.to_s)}
        before :each do
          authenticated_header(request, responsible)

          delete 'destroy', {
              id: applicant.id.to_s
          }
        end
        it 'returns status code 404' do
          expect(response).to have_http_status(204)
        end
      end
    end

    describe 'GET /applicants/find_by_nud' do
      let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
      let!(:applicant) { create(:applicant) }
      let!(:housing_request) { create(:housing_request, main_applicant_id: applicant.id) }
      let!(:nud) { create(:nud, unique_number: "111101657788975B02", applicant: applicant, housing_request: housing_request) }
      before :each do
        authenticated_header(request, responsible)
      end

      it 'return existing applicant' do
        get :find_by_nud, nud: nud.unique_number
        expect(response).to have_http_status(200)
        expect(json['data']['id']).to eq applicant.id.to_s
      end

      it 'return created applicant' do
        get :find_by_nud, nud: '1110121021518G1104', birth_date: '1982-10-10'
        expect(response).to have_http_status(200)
        expect(json['data']['id']).not_to be_nil
        expect(json['data']['id']).not_to eq applicant.id.to_s
      end

      it 'return invalid birth_date error' do
        get :find_by_nud, nud: '1110121021518G1104', birth_date: '1992-05-10'
        expect(response).to have_http_status(422)
        expect(json['errors'].first['code']).to include('invalid_birth_date')
      end

      it 'return not found for invalid nud' do
        get :find_by_nud, nud: '000000000000000000', birth_date: '1991-05-10'
        expect(response).to have_http_status(404)
        expect(json['errors'].first['code']).to include('invalid_nud')
      end
    end

    describe 'with bear minimum for eligibility checking' do
      let!(:admin) { create(:administrator) }

      let!(:threshold1) do
        create(
            :threshold,
            type: 'PLA LM',
            thresholds_map: {
                'Paris et communes limitrophes': {
                    'thresholds_list': [
                        13_050,
                        21_272,
                        27_883,
                        30_521,
                        36_307,
                        40_859
                    ],
                    'another_person': 4552
                },
                'IDF hors Paris et Communes limitrophes': {
                    'thresholds_list': [
                        13_050,
                        21_272,
                        25_569,
                        28_075,
                        33_238,
                        37_401
                    ],
                    'another_person': 4166
                },
                'Autres régions': {
                    'thresholds_list': [
                        11_342,
                        16_525,
                        19_872,
                        22_111,
                        25_870,
                        29_155
                    ],
                    'another_person': 3252
                }
            }
        )
      end

      let!(:user1) {create(:user)}
      let!(:applicant1) do
        create(
            :applicant,
            sne_profile_creation_date: '2016-09-24',
            user_id: user1.id,
            amount_year_minus_two: 9787.24,
            monthly_incomes: {
                'Salaire': 1045.23,
                'APL': 321.62
            }
        )
      end
      let!(:request1) do
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant1,
          seeked_departments: ['34']
        )
      end

      let!(:user2) {create(:user)}
      let!(:applicant2) do
        create(
          :applicant,
          sne_profile_creation_date: '2017-09-24',
          user_id: user2.id,
          amount_year_minus_two: 9787.24,
          monthly_incomes: {
            'Salaire': 1045.23,
            'APL': 321.62
          }
        )
      end
      let!(:request2) do
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant2,
          seeked_departments: ['34']
        )
      end

      let!(:nud1) { create(:nud,housing_request: request1, unique_number: '1')} 
      let!(:nud2) { create(:nud,housing_request: request2, unique_number: '2')} 

      let(:group1) { create(:group, user_ids: [user1.id]) }

      let!(:offer1) do
        create(
          :housing_offer,
          rent_with_charges: 234.32 + 42.15 + 12.71,
          rent_amount: 234.32,
          rental_charges: 42.15,
          parking_charges: 12.71,
          surface: 29,
          typology: 'T2',
          threshold_type: 'PLA LM',
          offer_status: 'LOGEMENT PUBLIE',
          postal_code: '34000',
          threshold_subtype: 'IDF hors Paris et Communes limitrophes',
          min_effort_rate: 0,
          readers: ["g_#{group1.id}"]
        )
      end

      let!(:offer2) do
        create(
          :housing_offer,
          rent_with_charges: 434.32 + 53.12 + 22.43,
          rent_amount: 434.32,
          rental_charges: 53.12,
          parking_charges: 22.43,
          surface: 34,
          typology: 'T2',
          threshold_type: 'PLA LM',
          offer_status: 'LOGEMENT PUBLIE',
          postal_code: '34000',
          threshold_subtype: 'Autres régions',
          min_effort_rate: 0,
          readers: ["g_#{group1.id}"]
        )
      end

      let!(:offer3) do
        create(
          :housing_offer,
          rent_with_charges: 434.32 + 53.12 + 22.43,
          rent_amount: 434.32,
          rental_charges: 53.12,
          parking_charges: 22.43,
          surface: 34,
          typology: 'T2',
          threshold_type: 'PLA LM',
          offer_status: "EN COURS D'ATTRIBUTION",
          postal_code: '34000',
          threshold_subtype: 'Autres régions',
          min_effort_rate: 0,
          readers: ["g_#{group1.id}"]
        )
      end

      before do
        [request1, request2].each do |hr|
          hr.validate_all_required_document
        end
      end

      describe 'GET /applicants/:id/eligible_offers' do
        before :each do
          authenticated_header(request, admin)
          get 'eligible_offers', id: applicant1.id
        end

        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end

        it 'returns valid JSON' do
          data = json['data']
          expect(data.length).to eq(2)
        end
      end

      describe 'GET /applicants/:id/eligible_offers with housing_channel' do
        let!(:channel_dalo) { create :housing_channel, key: 'dalo' }
        let!(:channel_pp) { create :housing_channel, key: 'priority_public' }

        before :each do
          offer1.update(housing_channel_keys: ['dalo'])
          request1.update(housing_channel_keys: ['priority_public'])
          authenticated_header(request, admin)
          get 'eligible_offers', id: applicant1.id
        end

        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end

        it 'returns valid JSON' do
          data = json['data']
          expect(data.length).to eq 1
          expect(data.map { |d| d['id'] }).to eq [offer2.id.to_s]
        end
      end
    end

    describe 'update reference_tax_revenues on hr' do 
      let!(:superadministrateur) { create(:user, auth_roles: ['superadministrateur']) }
      let!(:applicant) {create(:applicant, amount_year_minus_two: 100)}
      let!(:coapplicant) {create(:applicant, amount_year_minus_two: 100)}
      let!(:coapplicant2) {create(:applicant, amount_year_minus_two: 100)}
      let!(:hr) {create(:housing_request, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'}], is_active: true)}

      it 'coapplicant' do
        authenticated_header(request, superadministrateur)
        patch 'update', {
          id: coapplicant.id.to_s,
          data: {
            attributes: { amount_year_minus_two: 5000}
          }
        }

        expect(response).to have_http_status 200
        expect(hr.reload.reference_tax_revenues).to eq(5200)
      end

    end

    describe 'POST /switch' do
      let!(:reader) { create :user, roles: ['Superadmin', 'Administrator'] }
      let!(:entreprise) { create :user, auth_roles: ['entreprise'] }
      let!(:target) { create :user, main_applicant_id: applicant.id }
      let!(:entity) { create :entity, readers: [entreprise.id.to_s], siret: '12345678901234' }
      let!(:form_level) { create :entity_form_level, entity_id: entity.id }
      let!(:applicant) { create :applicant, main_employer_siret: entity.siret, auto_set_user: false }
      let!(:unkown_applicant) { create :applicant, main_employer_siret: entity.siret, auto_set_user: false }

      it 'returns unauthorized error' do
        authenticated_header(request, reader)
        post 'switch', { id: applicant.id }

        expect(response).to have_http_status 401
        expect(json['message']).to eq 'Votre role doit être \'entreprise\' pour effectuer cette action.'
      end

      context 'as entreprise' do
        let!(:housing_request) { create :active_request, main_applicant_id: applicant.id }
        before do
          authenticated_header(request, entreprise)
        end

        it 'returns not_found error' do
          post 'switch', { id: unkown_applicant.id }
          expect(response).to have_http_status 404
          expect(json['message']).to eq 'Utilisateur target non trouvé.'
        end

        it 'requires entity_id param' do
          post 'switch', { id: applicant.id }
          expect(response).to have_http_status 406
          expect(json['errors'][0]['message']).to eq 'param is missing or the value is empty: entity_id'
        end

        it 'returns not_found error' do
          post 'switch', { id: applicant.id, entity_id: '123' }
          expect(response).to have_http_status 404
          expect(json['success']).to eq false
          expect(json['message']).to eq 'Entity not found'
        end

        context 'when no entity_form' do
          it 'returns not_found error' do
            post 'switch', { id: applicant.id, entity_id: entity.id }
            expect(response).to have_http_status 404
            expect(json['success']).to eq false
            expect(json['message']).to eq 'delegation not found'
          end
        end

        context 'when entity_form without delegation and/or delegation signed' do
          let!(:entity_form) do
            create(
              :entity_form,
              housing_request_id: housing_request.id,
              entity_form_level_id: form_level.id,
              delegation_signed: false
            )
          end

          it 'returns unauthorized error' do
            post 'switch', { id: applicant.id, entity_id: entity.id }
            expect(response).to have_http_status 401
            expect(json['success']).to eq false
            expect(json['message']).to eq 'delegation not available or not signed'
          end
        end

        context 'when entity_form with delegation and delegation signed' do
          let!(:form_level) { create :entity_form_level, entity_id: entity.id, delegation: true, level: 4 }
          let!(:entity_form) do
            create(
              :entity_form,
              housing_request_id: housing_request.id,
              entity_form_level_id: form_level.id,
            )
          end
          before do
            post 'switch', { id: applicant.id, entity_id: entity.id }
          end
          it 'succeeds' do
            expect(response).to have_http_status 200
          end
          it 'returns target infos' do
            expect(json["email"]).to eq target.email
            expect(json['jwt_token']).not_to be_empty
          end

          context 'find by email' do
            let!(:target_by_email) { create :user, email: 'test@test.com' }
            let!(:applicant_by_email) { create :applicant, main_employer_siret: entity.siret, email: 'test@test.com', auto_set_user: false }
            let!(:housing_request) { create :active_request, main_applicant_id: applicant_by_email.id }
            before do
              post 'switch', { id: applicant_by_email.id, entity_id: entity.id }
            end
            it 'succeeds' do
              expect(response).to have_http_status 200
            end
            it 'returns target infos' do
              expect(json['email']).to eq target_by_email.email
              expect(json['jwt_token']).not_to be_empty
            end
          end
        end
      end
    end
  end

  describe Api::Dmo::ApplicantsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
    TAGS_API_DMO_APPLICANTS= 'API_DMO_APPLICANTS'
    include JwtAuthHelper
    extend HermesSwaggerJsonParams
    include SwaggerJsonAttributes
    let!(:dc_channel) {create(:housing_channel, key: 'dc')}

    context 'as admin' do
      let!(:admin) {create(:administrator)}

      path '/api/dmo/applicants' do
        post 'Create applicant' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'
          consumes 'application/json'
          security [Bearer: {}]
          json_params = {}
          MUTATE_PARAMS.each do |key, val|
            json_params[key]= {type: nice_type(val)}
          end
          json_params[:housing_request_ids]={type: :array}
          json_params[:entity_ids]={type: :array}
          parameter(
            name: :params,
            in: :body,
            schema: jsonify_params(json_params),
            description: '<span>' \
              '<div class="opblock opblock-delete" style="padding: 4px; align-items: center; display: flex">' \
              '<span class="opblock-summary-method" style="margin-right: 4px">WARNING</span>' \
              '<span class="no-style">Setting `main_employer_siret, will ' \
              ' change its relationship with the corresponding entity.</span></div></span>'
          )
          #jsonapi_params( **json_params )

          response '201', 'applicant created' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:params) do
              jsonapi_attributes(
                  user_id: admin.id.to_s
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'

            after do |_example|
              json = JSON.parse(response.body)

              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'applicants'
            end
          end
        end
      end

      path '/api/dmo/applicants' do
        get 'Retrieves all applicants' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'

          parameter name: 'ids[]', in: :query, type: :array, items: { type: :string }, collectionFormat: :multi, required: false
          ACCEPTABLE_PARAMS.each do |acceptable_param|
            type = :string
            if MUTATE_PARAMS[acceptable_param]
              type = nice_type(MUTATE_PARAMS[acceptable_param] )
            end
            parameter name: acceptable_param.to_s, in: :query, type: type, required: false
          end
          #VALID_OPERATORS.each do |valid_operator|
           # parameter( name: valid_operator.to_s, in: :query,type: :object, required: false)
          ## end
          security [Bearer: {}]
          paginator

          response '200', 'applicants found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:applicants) { create_list(:applicant, 5,user_id: admin.id.to_s) }

            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) { applicants }
            end
          end
        end
      end

      path '/api/dmo/applicants/find_by_nud' do
        get 'Retrieves or create applicant with nud' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'

          parameter name: :nud, in: :query, type: :string, required: true
          parameter name: :birth_date, in: :query, type: :string, required: false, description: "Doit être renseigné si le demandeur n'existe pas. Format attendu : YYYY-MM-DD"
          security [Bearer: {}]

          response '200', 'applicant found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:applicant) { create(:applicant) }
            let!(:housing_request) { create(:housing_request, main_applicant_id: applicant.id) }
            let!(:nud_inst) { create(:nud, unique_number: "111101657788975B02", applicant: applicant, housing_request: housing_request) }
            let!(:nud) { nud_inst.unique_number }
            include_context 'with integration test'
            include_context 'valid jsonapi ressource' do
              let(:expected_ressource) { applicant }
            end
          end

          response '200', 'applicant found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:nud) { '1110121021518G1104' }
            let!(:birth_date) { '1982-10-10' }
            include_context 'with integration test'
            run_test!
          end

          response '404', 'invalid nud' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:nud) { '000000000000000000' }
            let!(:birth_date) { '1992-05-10' }
            run_test!
          end

          response '422', 'invalid birth_date' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:nud) { '1110121021518G1104' }
            let!(:birth_date) { '1992-05-10' }
            run_test!
          end
        end
      end

      path '/api/dmo/applicants/{id}/eligible_offers' do
        get 'Retrieves eligible offers for a given applicant' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string,  required: true
          parameter name: 'ids[]', in: :query, type: :array, items: { type: :string }, collectionFormat: :multi, required: false
          security [Bearer: {}]
          OFFERS_ACCEPTABLE_PARAMS.each do |acceptable_param|
            parameter name: acceptable_param.to_s, in: :query,  required: false
          end
          paginator

          response '200', 'housing_offers found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let(:applicant) { create(:applicant,user_id: admin.id.to_s) }
            let(:id) { applicant.id }

            include_context 'not containings errors'
            run_test!

          end
        end
      end

      path '/api/dmo/applicants/{id}' do
        get 'Retrieves a single applicant by id' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]

          response '200', 'applicant found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let(:applicant) { create(:applicant) }
            let(:id) { applicant.id }

            include_context 'with integration test'
            include_context 'valid jsonapi ressource' do
              let(:expected_ressource) { applicant }
            end
          end
        end
      end

      path '/api/dmo/applicants/{id}' do
        put 'edit applicant' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'
          consumes 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          json_params = {}
          MUTATE_PARAMS.each do |key, val|
            json_params[key]= {type: nice_type(val)}
          end
          json_params[:main_applicant_id]={type: :string}
          json_params[:coapplicants]={type: :array}
          json_params[:housing_application_ids]={type: :array}
          json_params[:applicant_ids]={type: :array}
          json_params[:seeked_departments]={type: :array}
          json_params[:dependents]={type: :array}
          parameter(
            name: :params,
            in: :body,
            schema: jsonify_params(json_params),
            description:  '<span>' \
            '<div class="opblock opblock-delete" style="padding: 4px; align-items: center; display: flex">' \
            '<span class="opblock-summary-method" style="margin-right: 4px">WARNING</span>' \
            '<span class="no-style">Setting `main_employer_siret, will ' \
            ' change its relationship with the corresponding entity.</span></div></span>'
          )
          #jsonapi_params( **json_params )

          response '200', 'applicant created' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:applicant) {create(:applicant, user_id: admin.id.to_s)}
            let(:id) {applicant.id}
            let(:params) do
              jsonapi_attributes(
                  id: id,
                  first_name: 'Joseph',
                  last_name: 'Joestar'
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'

            after do |_example|
              json = JSON.parse(response.body)

              data = json['data']
              expect(data).to be_a Hash
              expect(data['id']).to_not be nil
              attributes = data['attributes']
              expect(attributes['first_name']).to eq 'Joseph'
              expect(attributes['last_name']).to eq 'Joestar'
            end
          end
        end
      end

      path '/api/dmo/applicants/{id}' do
        delete 'Delete a single applicant by id' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]

          response '204', 'applicant deleted' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let(:id) { create(:applicant, responsible: admin.id).id }
            run_test!
          end
        end
      end
    end

    context 'as entreprise' do
      let!(:entreprise) { create :user, auth_roles: ['entreprise'] }
      path '/api/dmo/applicants/{id}/switch' do
        post 'Create applicant' do
          tags TAGS_API_DMO_APPLICANTS
          produces 'application/json'
          consumes 'application/json'
          security [Bearer: {}]
          parameter name: :id, in: :path, type: :string
          parameter name: :params, in: :body, schema: {
            entity_id: { type: :string, required: true }
          }
          response '200', 'applicant switched' do
            let(:Authorization) {"Bearer #{token_for(entreprise)}"}
            let!(:entity) { create :entity, readers: [entreprise.id.to_s], siret: '12345678901234' }
            let!(:applicant) { create :applicant, main_employer_siret: entity.siret, auto_set_user: false }
            let!(:user) { create :user, email: 'test@test.com', main_applicant_id: applicant.id }
            let!(:housing_request) { create :active_request, main_applicant_id: applicant.id }
            let!(:form_level) { create :entity_form_level, entity_id: entity.id, delegation: true, level: 4 }
            let!(:entity_form) do
              create(
                :entity_form,
                housing_request_id: housing_request.id,
                entity_form_level_id: form_level.id,
              )
            end
            let(:id) { applicant.id }
            let(:params) do
              { entity_id: entity.id }
            end
            include_context 'with integration test'
            include_context 'not containings errors'

            after do |_example|
              json = JSON.parse(response.body)

              expect(json['email']).to eq 'test@test.com'
              expect(json["jwt_token"]).not_to be_empty
            end
          end
        end
      end
    end
  end
end