require 'rails_helper'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  include Rack::Test::Methods
  
  describe Api::Dmo::NotificationCenter::SubscriptionsController, :type => :api do
    
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper
    let!(:user) { create(:user) }
    let!(:applicant) { create(:applicant, notifications_count: 5) }
    let!(:new_key) { 'testing.key' }

    describe 'Actions' do
      before :each do
        authenticated_header(request, user)
      end

      it 'return subscription' do
        get :show, targetable_type: 'Hermes::Applicant', id: applicant.id, subscription_id: applicant.subscription.id

        expect(response).to have_http_status(200)
        expect(json['data']['id']).to eq(applicant.subscription.id.to_s)
      end

      it 'should subscribe to a new key' do
        put :subscribe, targetable_type: 'Hermes::Applicant', id: applicant.id, subscription_id: applicant.subscription.id, key: new_key

        expect(response).to have_http_status(200)
        expect(applicant.subscription.subscribed_to_email?(new_key)).to be(true)
      end

      it 'should unsubscribe to an existing key' do
        applicant.subscription.subscribe_to_email(new_key)
        put :unsubscribe, targetable_type: 'Hermes::Applicant', id: applicant.id, subscription_id: applicant.subscription.id, key: new_key

        expect(response).to have_http_status(200)
        expect(applicant.subscription.subscribed_to_email?(new_key)).to be(false)
      end

      it 'should return error if provider does not exist' do
        put :unsubscribe, targetable_type: 'Hermes::Applicant', id: applicant.id, subscription_id: applicant.subscription.id, key: new_key, provider: :hello_word

        expect(response).to have_http_status(404)
      end
    end
  end
end