require 'rails_helper'
# require Hermes::Engine.root
#                       .join(*%w[spec factories lessors_factory])
#                       .to_s
# require Hermes::Engine.root
#                       .join(*%w[spec factories lessor_objectives_factory])
#                       .to_s
# require Hermes::Engine.root
#                       .join(*%w[spec factories housing_offers_factory])
#                       .to_s

RSpec.describe Hermes::LessorObjective, type: :model do
  it 'has a valid factory' do
    expect(FactoryBot.build(:lessor_objective)).to be_valid
  end

  # needed while old transition == true
  let!(:housing_family) { create(:family, slug: Hermes::ENV['offer_family_slug']) }

  let!(:base_lessor) { create(:lessor) }

  it 'should have created base_lessor' do
    expect(base_lessor).to be_present
  end

=begin
  let!(:lessor_objective_all) do
    create(
      :lessor_objective,
      department: '34',
      district: 'Montpellier',
      threshold_type: 'PLUS 60%',
      priority_public: true,
      new_offer: false,
      target_by_typologies: { all: 24 },
      realized_by_typologies: { all: 42 },
      lessor_id: base_lessor.id
    )
  end
=end

  # let!(:offer_test) do
  #   create(
  #     :tagset,
  #     superfields: {

  #     }
  #   )
  # end

  describe 'objectives are linked to lessor' do
    let!(:requalified_offer) do
      create(
        :housing_offer,
        department: '34',
        district: 'Montpellier',
        threshold_type: 'PLUS 60%',
        priority_public: true,
        new_offer: false,
        typology: 'T4',
        lessor_id: base_lessor.id,
        offer_status: 'requalification'
      )
    end

    let!(:offers_batch1) do
      create_list(
        :housing_offer,
        3,
        department: '34',
        district: 'Montpellier',
        threshold_type: 'PLUS 60%',
        priority_public: true,
        new_offer: false,
        typology: 'T4',
        lessor_id: base_lessor.id
      )
    end

    let!(:lessor_objective1) do
      create(
          :lessor_objective,
          department: '34',
          district: 'Montpellier',
          threshold_type: 'PLUS 60%',
          priority_public: true,
          new_offer: false,
          target_by_typologies: { all: 24 },
          realized_by_typologies: { all: 42 },
          lessor_id: base_lessor.id,
          activation_date: DateTime.new(DateTime.current.year, 5, 5).utc
      )
    end

    let!(:lessor_objective2) do
      create(
          :lessor_objective,
          department: '44',
          district: 'Montpellier',
          threshold_type: 'PLUS 60%',
          priority_public: true,
          new_offer: false,
          target_by_typologies: { all: 24 },
          realized_by_typologies: { all: 42 },
          lessor_id: base_lessor.id,
          activation_date: DateTime.new(DateTime.current.year, 1, 1).utc
      )
    end
    let!(:lessor_objective3) do
      create(
        :lessor_objective,
        department: '412',
        district: 'Mpellier',
        threshold_type: 'PL',
        priority_public: false,
        new_offer: false,
        target_by_typologies: { all: 24 },
        lessor_id: base_lessor.id,
        activation_date: DateTime.new(DateTime.current.year, 1, 1).utc
      )
    end
    let!(:offers_batch3) do
      create_list(
        :housing_offer,
        4,
        department: '95',
        district: 'Paris',
        threshold_type: 'PLUS',
        priority_public: false,
        new_offer: true,
        typology: 'T2',
        lessor_id: base_lessor.id
      )
    end

    it 'shows update_objective_difference_by_typologies' do
      expect(lessor_objective3.objective_difference_by_typologies).to eq({"all"=>0})
    end

    it 'should return the right number of relations' do
      expect(Hermes::HousingOffer.where(lessor_id: base_lessor.id).count).to be 8
      expect(base_lessor.reload.housing_offers.count).to eq(8)
    end

    it 'should not include requalified offers' do
      [lessor_objective1, lessor_objective2, lessor_objective3].each do |objective|
        expect(Hermes::HousingOffer.will_trigger_objective(objective.id)).not_to include(requalified_offer)
      end
    end
  end

  describe "#increment_realized_for" do
    let(:housing_offer) { create :housing_offer }
    let(:lessor_objective) do
      create(
        :lessor_objective,
        target_by_typologies: { T2: 3 },
        realized_by_typologies: { T2: 0 },
        lessor_id: base_lessor.id,
        activation_date: DateTime.new(DateTime.current.year, 5, 5).utc
      )
    end
    let(:histories) { lessor_objective.lessor_objective_histories }

    subject { lessor_objective.increment_realized_for('T2', housing_offer) }

    it "increments realized objective for a typology" do
      expect { subject }.to change { lessor_objective.reload.realized_by_typologies[:T2] }.from(0).to 1
    end

    it "creates a new lessor_objective_history" do
      subject
      expect(histories.count).to eq 1
      expect(histories.last.reload.target_by_typologies).to eq ({ 'T2' => 3 })
      expect(histories.last.reload.realized_by_typologies).to eq ({ 'T2' => 1 })
      expect(histories.last.reload.housing_offer_id).to eq housing_offer.id
    end
  end
end
